import json

from database_mysql_local.connector import Connector
from database_mysql_local.generic_crud import GenericCRUD
from logger_local.LoggerLocal import Logger
from logger_local.MetaLogger import module_wrapper

from .Constants import (COMMUNICATION_TYPE, DIALOG_WORKFLOW_CODE_LOGGER_OBJECT,
                        CommunicationTypeEnum, WorkflowActionEnum, seperator)

logger = Logger.create_logger(object=DIALOG_WORKFLOW_CODE_LOGGER_OBJECT)

# from circles_local_aws_s3_storage_python.AWSStorage import AwsS3Storage

# TODO:Let's move to Object Oriented Programming
DEFAULT_STATE_ID = 1


def process_message(communication_type: CommunicationTypeEnum, action_type: WorkflowActionEnum,
                    message: str) -> str or json:
    """Processes message according to which communication_type is used: console or websocket
        If console then we continue the code normally.
        If websocket then we create a json out of the given message and exit(0)"""

    if communication_type == CommunicationTypeEnum.CONSOLE:
        message = message.replace(seperator, '\n')
    elif communication_type == CommunicationTypeEnum.WEBSOCKET:
        message = update_json_message(action_type, message)

    return message


def update_json_message(action_type: WorkflowActionEnum, message: str) -> json:
    """Update message to json format"""

    dict_message = {"message": message}
    if action_type == WorkflowActionEnum.AGE_DETECTION:
        dict_message["type"] = "command"
    else:
        dict_message["type"] = "text"
    json_message = json.dumps(dict_message)

    return json_message


# TODO get_last_dialog_workflow_state_id() - This is the name in the database, let's be consistance
# TODO Shall we move this function as static method in DialogWorkflowLocal?
# TODO Change the order of the parameters (alphabet order): campaign, channel, message, profile_context
# TODO Change the algorithem
# 1st priority profile_id+ message_id
# 2nd priority profile_id+ channel_id
# 3rd priority profile_id+ campaign_id
def get_curr_state_id(profile_id: int, channel_id: int = None, campaign_id: int = None, message_id: int = None) -> int:
    """Returns profiles' curernt state number"""

    crud = GenericCRUD(default_schema_name='campaign_message_profile',
                       default_view_table_name='campaign_message_profile_general_view')
    select_clause_value = "last_dialog_workflow_state_id"
    where = "profile_id = %s"
    params = (profile_id,)
    if channel_id:
        where += " AND channel_id = %s"
        params += (channel_id,)
    if campaign_id:
        where += " AND campaign_id = %s"
        params += (campaign_id,)
    if message_id:
        where += " AND message_id = %s"
        params += (message_id,)
    # Note: there can be multiple matchs
    curr_state = crud.select_one_value_by_where(where=where, params=params, select_clause_value=select_clause_value,
                                                order_by="updated_timestamp DESC")
    if not curr_state:
        if campaign_id:
            logger.warning(f"No state found for the given profile_id, looking for state by {campaign_id=}",
                           object=locals())
            curr_state = get_state_id_by_campaign_id(campaign_id)
        if not curr_state:
            curr_state = DEFAULT_STATE_ID
            logger.warning("No state found for the given campaign_id, using default state " + str(DEFAULT_STATE_ID),
                           object=locals())
            crud.insert(data_dict={"profile_id": profile_id, "campaign_id": campaign_id,
                                   "message_id": message_id, "last_dialog_workflow_state_id": curr_state})

    logger.info(f"Current state: {profile_id=}, {curr_state=}", object=locals())
    return curr_state


def get_state_id_by_campaign_id(campaign_id: int) -> int:
    crud = GenericCRUD(default_schema_name='campaign', default_view_table_name='campaign_general_view')
    start_state_id = crud.select_one_value_by_column_and_value(
        select_clause_value="start_state_id", column_name="campaign_id", column_value=campaign_id)

    return start_state_id


def update_profile_curr_state_in_db(new_state: int, profile_id: int, channel_id: int = None,
                                    campaign_id: int = None, message_id: int = None) -> None:
    """This function updates the last_dialog_workflow_state_id in campaign_message_profile to the new state.
       Note that this function UPDATES the field and doesn't INSERT into the table. """

    crud = GenericCRUD(default_schema_name='campaign_message_profile')
    where = "profile_id = %s"
    params = (profile_id,)
    if channel_id:
        where += " AND channel_id = %s"
        params += (channel_id,)
    crud.update_by_where(where=where, params=params, data_dict={"last_dialog_workflow_state_id": new_state},
                         limit=1, order_by="updated_timestamp DESC")
    if message_id and campaign_id:
        crud.update_by_column_and_value(schema_name="message", table_name="message_table",
                                        column_name="message_id", column_value=message_id,
                                        data_dict={"campaign_id": campaign_id})


# TODO: Make sure we have good testing coverage to this function
def store_age_detection_picture(age_range: str, profile_id: int) -> None:
    """Stores the picture in Nir's storage schema, gets a storage_id and inserts into the computer_vision_storage_table"""
    # storage = AwsS3Storage(bucket_name="storage.us-east-1.dvlp1.bubblez.life", region="us-east-1")
    # storage_id = storage.upload_file("C:\\Users\\User\\OneDrive\\Circles\\age-detection-backend\\src\\alonPicture.png", "Alon's picture", "", 1)

    # age_range_split = age_range.split('-')
    # TODO Why this is commented?
    # TODO Why we commented this? Let's uncomment
    # min_age = int(age_range_split[0][:len(age_range_split[0])-1])
    # max_age = int((age_range_split[1])[1:])
    # cursor.execute("""USE computer_vision_storage""")
    # cursor.execute("""INSERT INTO computer_vision_storage_table
    #                 (storage_id, profile_id, min_age, max_age) VALUES (%s, %s, %s, %s)""",
    #                [storage_id, profile_id, min_age, max_age])
    # connection.commit()
    pass


def get_child_nodes_of_current_state(fields: list, table_name: str, values_from_where_to_select: tuple,
                                     variables_from_where_to_select: list) -> list[dict]:
    """Recieves all the relevant information and selects the child nodes of the current state from the given table, and returns them"""
    crud = GenericCRUD(default_schema_name='dialog_workflow', default_table_name=table_name)
    select_clause_value = ", ".join(fields)
    where = " AND ".join(f"{variable} = %s" for variable in variables_from_where_to_select)
    child_nodes = crud.select_multi_dict_by_where(select_clause_value=select_clause_value,
                                                  where=where, params=values_from_where_to_select)
    # cursor.execute(sql_query, values_from_where_to_select)
    # child_nodes = cursor.fetchall()

    return child_nodes


class Group:
    def __init__(self, parameter1: int) -> None:
        self.parameter1 = parameter1

    # TODO Make sure we have tests for this and all other methods        
    def get_group_childs_id_bellow_parent_id(self) -> list[int]:
        """returns all the childs ids below the given parent_id.
            This function gets all the id's of records that their parent_state_id is the given id, and continues to add id's  recursively 
            until all the records that their parent_state_id matches an id in the table."""

        connection = Connector.connect("group")
        cursor = connection.cursor(dictionary=True, buffered=True)
        # TODO Fix this SQL  [why?]
        cursor.execute("""
        WITH RECURSIVE cte AS (
            SELECT group_view.group_id FROM group_view WHERE group_id = %s
            UNION ALL
            SELECT group_view.group_id FROM group_view 
            JOIN cte ON group_view.parent_group_id = cte.group_id
        )
        SELECT cte.group_id FROM cte""", (self.parameter1,))
        group_id_dict = cursor.fetchall()
        group_childs_id = [group['group_id'] for group in group_id_dict]

        return group_childs_id

    def get_child_group_names(self) -> list:
        """Gets all the child title names from the ml table of the given parent_id."""

        group_ids = self.get_group_childs_id_bellow_parent_id()
        crud = GenericCRUD(default_schema_name='group', default_view_table_name='group_ml_view')
        select_clause_value = "title"
        group_id_placeholders = ','.join(["%s"] * len(group_ids))
        params = tuple(group_ids)
        where = f"group_id IN ({group_id_placeholders})"
        child_group_names = crud.select_multi_value_by_where(select_clause_value=select_clause_value,
                                                             where=where, params=params)
        # child_group_names = [group['title'] for group in group_name_dict]

        return child_group_names

    # def get_child_group_id(self) -> list[int]:
    #     """Gets the id of all the records with the given group name"""
    #     cursor.execute("""USE `group`""")
    #     cursor.execute("""SELECT group_id from group_ml_en_view WHERE title = %s""", [self.parameter1])
    #     group_id_dict = cursor.fetchall()
    #     return [group['group_id'] for group in group_id_dict]        


# TODO Shall we move this to a menu class?
def generic_menu(*, options: list, got_response: bool, chosen_numbers: str, choose_one_option: bool,
                 outgoing_message: str) -> list[int] or json:
    """A generic function for displaying a menu for the user.
        Returns: If not got_response: the menu options as json to send to user.
                 Otherwise, returns the chosen numbers as list in int"""
    logger.info(object=locals())
    outgoing_message_json = None
    if not got_response:
        outgoing_message += f"Please choose EXACTLY ONE option between 1-{len(options)}:{seperator}" if choose_one_option else \
            f"Please select your desired choices, You may select any of the numbers between 1-{len(options)} with a comma seperator between each choice:{seperator}"
        outgoing_message_json = None
        for i, option in enumerate(options):
            outgoing_message = outgoing_message + f'{i + 1}) {option}{seperator}'
            outgoing_message_json = process_message(communication_type=COMMUNICATION_TYPE,
                                                    action_type=WorkflowActionEnum.TEXT_MESSAGE_ACTION,
                                                    message=outgoing_message)
        if COMMUNICATION_TYPE == CommunicationTypeEnum.WEBSOCKET:
            return outgoing_message_json

    if not chosen_numbers and COMMUNICATION_TYPE == CommunicationTypeEnum.CONSOLE:
        if outgoing_message_json:  # print to console
            print(outgoing_message_json.replace(seperator, "\n"))
        chosen_numbers = input(f"Please choose the options you want between 1-{len(options)}: ")
    chosen_numbers = chosen_numbers.split(',')
    generic_menu_result = [int(x) for x in chosen_numbers if x.isdigit()]
    if not generic_menu_result:
        logger.exception("No chosen numbers", object=locals())
    elif not all(x.isdigit() for x in chosen_numbers):
        logger.warning("Not all chosen numbers are digits", object=locals())
    return generic_menu_result


module_wrapper(logger)



import json
import random

from database_mysql_local.generic_crud import GenericCRUD
from language_remote.lang_code import LangCode
from logger_local.MetaLogger import MetaLogger
from message_local.CompoundMessage import CompoundMessage
from message_local.Recipient import Recipient
from python_sdk_remote.utilities import to_json
from user_context_remote.user_context import UserContext

from .Constants import DIALOG_WORKFLOW_CODE_LOGGER_OBJECT, WorkflowActionEnum
from .ProfileContext import DialogWorkflowRecord
from .action import Action
from .utils import get_curr_state_id, update_profile_curr_state_in_db

compound_message_keys = ["form_id", "form_page_number", "index_number", "message_template_id",
                         "message_template_text_block_id", "question_id", "variable_id", "variable_value"]

user_context = UserContext()  # Should be called only once, and may be updated later with jwt in serverless


class DialogWorkflowLocal(GenericCRUD, metaclass=MetaLogger, object=DIALOG_WORKFLOW_CODE_LOGGER_OBJECT):
    def __init__(self, *, is_test_data: bool = False) -> None:
        GenericCRUD.__init__(self, default_schema_name='dialog_workflow',
                             default_table_name='dialog_workflow_state_table',
                             default_view_table_name='dialog_workflow_state_view',
                             is_test_data=is_test_data)

    # TODO Please use MessagesLocal to store the message(s) recieved and anwers
    # TODO Please use the function defined in avatar similar to get_dialog_workflow_avatar_profile_id( profile_id1, prompt, group_id, profile_id2...) using avatar_table and avtar_group_table
    # TODO profile_curr_state_id
    def get_dialog_workflow_record(self, *, profile_curr_state: int, lang_code: LangCode,
                                   incoming_message: str = None) -> DialogWorkflowRecord:
        """Get all potential records in a specific state and choose randomly one of them"""
        select_clause_value = ("state_id, parent_state_id, workflow_action_id, lang_code, parameter1, variable1_id, "
                               "result_logical, result_figure_min, result_figure_max, next_state_id, "
                               "no_feedback_milliseconds, next_state_id_if_there_is_no_feedback, message_template_id")
        where = "state_id = %s AND (lang_code = %s OR lang_code IS NULL)"
        if incoming_message and not incoming_message.isdigit() and not all(
                x.isdigit() for x in incoming_message.split(',')):
            # can't be WorkflowActionEnum.MENU_ACTION, as it is expecting number(s)
            where += " AND workflow_action_id <> " + str(WorkflowActionEnum.MENU_ACTION.value)
        optional_records = self.select_multi_dict_by_where(
            where=where, params=(profile_curr_state, lang_code.value), select_clause_value=select_clause_value)
        if not optional_records:
            # TODO: send the client multiple languages and let it choose.
            raise Exception(f"No records found for state_id: {profile_curr_state} and language: {lang_code}")
        random_record = random.choice(optional_records)  # TODO: is it the right logic?
        if len(optional_records) > 1:
            self.logger.info("Choosed random record", object={
                "incoming_message": incoming_message, "random_record": random_record,
                "optional_records": optional_records})
        else:
            self.logger.info("Selected record", object={
                "incoming_message": incoming_message, "random_record": random_record})
        dialog_workflow_record = DialogWorkflowRecord(random_record)
        return dialog_workflow_record

    # TODO: get channel form campaign
    def post_message_and_get_message_json(self, *, incoming_message: str, channel_id: int = None,
                                          campaign_id: int = None, profile_id: int = None,
                                          previous_message_id: int = None) -> json:
        """This function is supposed to serve as a POST request using REST API.
        It runs until needing input from the user, which it then sends a JSON to the user with the message and exits
        param: incoming_message: the message to send"""
        # TODO: If it is an answer / forum response, write the answer to profile_question, and retrieve if already answered
        # TODO: get default variable for form - probably we have function
        # skip question if we have answer and it's not a form - move to next state
        incoming_message = incoming_message or ""
        self._save_compound_message_answer(incoming_message)
        profile_id = profile_id or user_context.get_effective_profile_id()
        lang_code = user_context.get_effective_profile_preferred_lang_code()
        # TODO Save the message using MessagesLocal (from DIALOG_WORKFLOW_PROFILE_ID, to UserContext.getEffectiveProfileId, ...)
        # TODO profile_curr_state_id
        profile_curr_state = get_curr_state_id(profile_id=profile_id, channel_id=channel_id,
                                            campaign_id=campaign_id, message_id=previous_message_id)
        got_response = incoming_message.strip() != ""  # This variable indicates if we must act now as we got a response from the user or as if we should send one to him
        init_action = Action(incoming_message=incoming_message, profile_id=profile_id,
                             lang_code=lang_code, profile_curr_state_id=profile_curr_state)
        message_id = previous_message_id
        while True:
            dialog_workflow_record: DialogWorkflowRecord = self.get_dialog_workflow_record(
                profile_curr_state=init_action.profile_curr_state_id, lang_code=lang_code,
                incoming_message=incoming_message)
            self.logger.info("dialog_workflow_record", object=dialog_workflow_record.__dict__)
            selected_act = init_action.act(dialog_workflow_record, got_response)
            outgoing_message = selected_act.get("outgoing_message")
            # TODO Save the message using MessagesLocal (from DIALOG_WORKFLOW_PROFILE_ID, to UserContext.getEffectiveProfileId, ...)
            outgoing_compound_message_json = None
            if outgoing_message is not None:
                if isinstance(outgoing_message, dict):  # is form
                    outgoing_message["current_state"] = init_action.profile_curr_state_id
                    outgoing_message["next_state"] = dialog_workflow_record.next_state_id
                    outgoing_compound_message_json = to_json(outgoing_message)
                elif isinstance(outgoing_message, list):  # is menu
                    # TODO "current_state_id", "next_state_id" 
                    outgoing_message.append({"current_state": init_action.profile_curr_state_id,
                                             "next_state": dialog_workflow_record.next_state_id})
                    outgoing_compound_message_json = to_json(outgoing_message)
                else:
                    recipient = Recipient(profile_id=profile_id, preferred_lang_code_str=lang_code.value,
                                          user_id=user_context.get_effective_user_id(),
                                          first_name=user_context.get_real_name())
                    compound_message = CompoundMessage(original_body=outgoing_message, recipients=[recipient])
                    outgoing_compound_message_dict = compound_message.get_compound_message_dict()
                    outgoing_compound_message_dict["current_state"] = init_action.profile_curr_state_id
                    outgoing_compound_message_dict["next_state"] = dialog_workflow_record.next_state_id
                    outgoing_compound_message_json = to_json(outgoing_compound_message_dict)
                    # TODO: compound_message.get_message_id()
                    message_id = compound_message.message_ids[0] if compound_message.message_ids else None

            if not selected_act.get("is_state_changed"):
                init_action.profile_curr_state_id = dialog_workflow_record.next_state_id
            update_profile_curr_state_in_db(new_state=init_action.profile_curr_state_id, profile_id=profile_id,
                                            channel_id=channel_id, campaign_id=campaign_id, message_id=message_id)
            if outgoing_compound_message_json:
                return outgoing_compound_message_json
            got_response = False

    def _save_compound_message_answer(self, incoming_message: str) -> None:
        """Save the answer to the form in the database"""
        # interface CompoundMessageAnswer {
        #   form_id: number;
        #   form_page_number: number;
        #   index_number: number;
        #   message_template_id: number;
        #   message_template_text_block_id: number;
        #   question_id: number;
        #   variable_id: number;
        #   variable_value: number | string | boolean;
        # }
        # Determine if the incoming message is a compound message answers:
        if incoming_message.startswith("[") and incoming_message.endswith("]"):
            try:
                compound_message_answers = json.loads(incoming_message)
                if not compound_message_answers or not isinstance(compound_message_answers, list) or not all(
                        key in compound_message_answers[0] for key in compound_message_keys):
                    return
            except Exception:
                return
        else:
            return

        for compound_message_answer in compound_message_answers:
            data_dict = {
                "user_id": compound_message_answer.get("user_id") or user_context.get_effective_user_id(),
                "profile_id": compound_message_answer.get("profile_id") or user_context.get_effective_profile_id(),
            }
            if compound_message_answer["variable_value"].isdigit():
                data_dict["answer_number"] = int(compound_message_answer["variable_value"])
            else:
                data_dict["answer_text"] = compound_message_answer["variable_value"]
            if compound_message_answer["question_id"]:
                data_dict["question_id"] = compound_message_answer["question_id"]
                self.insert(schema_name="profile_question", table_name="profile_question_table",
                            data_dict=data_dict)
            if compound_message_answer["variable_id"]:
                data_dict["variable_id"] = compound_message_answer["variable_id"]
                self.insert(schema_name="profile_variable", table_name="profile_variable_table",
                            data_dict=data_dict)

        self.logger.info("Saved compound_message_answers", object={
            "compound_message_answers": compound_message_answers})



# import msvcrt
# import requests
# from AgeDetection import DetectAge
from database_mysql_local.connector import Connector  # TODO: use crud
from language_remote.lang_code import LangCode
from logger_local.MetaLogger import MetaLogger
from message_local.CompoundMessage import CompoundMessage
from message_local.Recipient import Recipient
from user_context_remote.user_context import UserContext
from variable_local.template import ReplaceFieldsWithValues
from variable_local.variables_local import VariablesLocal

from src.Constants import (COMMUNICATION_TYPE, DIALOG_WORKFLOW_CODE_LOGGER_OBJECT,
                           CommunicationTypeEnum, WorkflowActionEnum, seperator)
from src.ProfileContext import DialogWorkflowRecord, ProfileContext
from src.utils import (Group, generic_menu, get_child_nodes_of_current_state,
                       get_curr_state_id, process_message,
                       update_profile_curr_state_in_db)


class Action(metaclass=MetaLogger, object=DIALOG_WORKFLOW_CODE_LOGGER_OBJECT):
    def __init__(self, *, incoming_message: str, profile_id: int,
                 lang_code: LangCode, profile_curr_state_id: int,
                 variables: VariablesLocal = None) -> None:
        self.incoming_message = incoming_message
        self.profile_id = profile_id
        self.lang_code = lang_code
        self.variables = variables or VariablesLocal()
        self.profile_curr_state_id = profile_curr_state_id
        self.accumulated_message = ""
        self.profile_context = ProfileContext(profile_id=self.profile_id, curr_state_id=profile_curr_state_id)
        self.record: DialogWorkflowRecord or None = None  # type: ignore
        self.got_response = False
        self.user_context = UserContext()

    def act(self, dialog_workflow_record: DialogWorkflowRecord, got_response: bool) -> dict:
        """This function applies the action of the relevant record with the profile's current state id.
        Params:
            1. dialog_workflow_record: The current record from the dialog workflow state table that is applied.
            2. got_response a bool indicator telling us if the user had sent back a response to the last outgoing message
               that we sent him from the same action or not. This will help understand if the action should
               apply the begging part of the action (i.e. to send a request to the user),
               or the second part of the action (i.e. got a response and now deal with it)

        Returns:
            1. True if the action resulted in a change of state, False otherwise
            2. The outgoing message json object that the user will recieve if the action needed to send a message to the user,
                or None if the action doesn't need to send a message to the user .

        Note : some of the actions are divided into 2 parts:
            1. The action before sendig the outgoing message to the user (i.e. got_response = false)
            2. The action after sending the outgoing message, and getting an incoming_message reply back (i.e. got_respone = true). """

        self.record = dialog_workflow_record
        self.got_response = got_response
        action = WorkflowActionEnum(self.record.workflow_action_id)
        self.logger.info(
            f"Applying action {action.name} ({action.value}) for profile_id {self.profile_id} in state_id {self.profile_curr_state_id}")
        if action == WorkflowActionEnum.LABEL_ACTION:
            selected_act = {"is_state_changed": False}
        elif action == WorkflowActionEnum.TEXT_MESSAGE_ACTION:
            selected_act = self.text_message_action()
        elif action == WorkflowActionEnum.QUESTION_ACTION:
            selected_act = self.question_action()
        elif action == WorkflowActionEnum.JUMP_ACTION:
            selected_act = self.jump_action()
        elif action == WorkflowActionEnum.SEND_REST_API_ACTION:
            selected_act = self.send_rest_api_action()
        elif action == WorkflowActionEnum.ASSIGN_VARIABLE_ACTION:
            selected_act = self.assign_variable_action()
        elif action == WorkflowActionEnum.INCREMENT_VARIABLE_ACTION:
            selected_act = self.increment_variable_action()
        elif action == WorkflowActionEnum.DECREMENT_VARIABLE_ACTION:
            selected_act = self.decrement_variable_action()
        elif action == WorkflowActionEnum.MENU_ACTION:
            selected_act = self.menu_action()
        elif action == WorkflowActionEnum.AGE_DETECTION:
            selected_act = self.age_detection()
        elif action == WorkflowActionEnum.MULTI_CHOICE_POLL:
            selected_act = self.multi_choice_poll()
        elif action == WorkflowActionEnum.PRESENT_CHILD_GROUPS_NAMES_BY_ID:
            selected_act = self.present_child_groups_names_by_id()
        elif action == WorkflowActionEnum.PRESENT_GROUPS_WITH_CERTAIN_TEXT:
            selected_act = self.present_groups_with_certain_text()
        elif action == WorkflowActionEnum.INSERT_MISSING_DATA:
            selected_act = self.insert_missing_data()
        elif action == WorkflowActionEnum.PRESENT_AND_CHOOSE_SCRIPT:
            selected_act = self.present_and_choose_script()
        elif action == WorkflowActionEnum.PRESENT_FORM:
            selected_act = self.present_form_action()
        elif action == WorkflowActionEnum.DISPLAY_MESSAGE_TEMPLATE:
            selected_act = self.display_message_template_action()
        else:
            raise ValueError(f"Action {action} is not supported")

        # TODO Please add support to question_schema i.e. ASK_QUESTION_ID
        return selected_act

    def text_message_action(self) -> dict:
        """Prints the paramter1 message after formatting:
        "Hello {First Name}, how are you {feeling|doing}?" --> "Hello Tal, how are you doing? """
        message = self.record.parameter1
        if not message:
            raise ValueError(f"Parameter1 is empty for record {self.record}")
        replace_fields_with_values_class = ReplaceFieldsWithValues(
            message=message, lang_code=self.lang_code, variables=self.variables)
        formatted_message = replace_fields_with_values_class.get_formatted_message(profile_id=self.profile_id)
        self.accumulated_message = self.accumulated_message + formatted_message + seperator
        outgoing_message = process_message(
            communication_type=COMMUNICATION_TYPE, action_type=WorkflowActionEnum.TEXT_MESSAGE_ACTION,
            message=self.accumulated_message)
        text_message_act = {"is_state_changed": False, "outgoing_message": outgoing_message}
        return text_message_act

    # TODO Rename all those functions to question_workflow_action() - Add the word 'workflow'
    # TODO Use question.question_table
    # TODO send to the logger the question_id
    # TODO confirm we didn't ask this question in the period mention in question.question_table
    def question_action(self) -> dict:
        """Asks a question and waits for an answer from user on STDIN. If the user responded in a certain amount of time,
        then moves to next state, otherwise moves to a different state.
        Note: this function waits for input for a certain amount of time only if using console application.
              If using websocket we send a json message to the user and exit the code normally"""

        if not self.got_response:
            self.accumulated_message += self.record.parameter1 or ""
            outgoing_message = process_message(
                communication_type=COMMUNICATION_TYPE, action_type=WorkflowActionEnum.TEXT_MESSAGE_ACTION,
                message=self.accumulated_message)
            if COMMUNICATION_TYPE == CommunicationTypeEnum.WEBSOCKET:
                question_act = {"outgoing_message": outgoing_message, "is_state_changed": False}
                return question_act
            else:
                self.accumulated_message = ""

            # waiting_time = self.record.no_feedback_milliseconds
            # start_time = time.monotonic()
            # input_str = None
            # while True:
            # if msvcrt.kbhit():
            # input_str = input().strip()
            # insert_profile_variable_value(self.profile_id, self.record.variable1_id, input_str, self.profile_curr_state_id)
            # break
            # elif time.monotonic() - start_time > waiting_time:
            # break
            input_str = input("Please insert your answer: ").strip()
            if input_str is None:
                self.profile_curr_state_id = self.record.next_state_id_if_there_is_no_feedback
                got_response = True
            else:
                got_response = False
            question_act = {"is_state_changed": got_response}
            return question_act
        else:
            self.variables.set_variable_value_by_variable_id(
                self.record.variable1_id, self.incoming_message, self.profile_id, self.profile_curr_state_id)
            question_act = {"is_state_changed": False}
            return question_act

    def jump_action(self) -> dict:
        """Jumps from one state to another."""
        self.profile_curr_state_id = int(self.record.parameter1)
        # TODO: do we also have to send channel_id and campaign_id?
        update_profile_curr_state_in_db(new_state=self.profile_curr_state_id, profile_id=self.profile_id)
        jump_action = {"is_state_changed": True}
        return jump_action

    @staticmethod
    def send_rest_api_action() -> dict:
        """Sends a REST API post"""
        # TODO: implament this function

        # api_url = self.record.parameter1
        # payload_variable_id = self.record.variable1_id
        # json_payload_string = self.variables.get_variable_value_by_variable_id(
        #             payload_variable_id, self.language, self.profile_id)
        # json_payload = json.loads(json_payload_string)
        # incoming_message = requests.post(api_url, json=json_payload)
        # incoming_message_string = json.dumps(incoming_message.json())
        # insert_profile_variable_value(self.profile_id, self.variable.get_variable_id("Post Result"), incoming_message_string, self.profile_curr_state_id)  # noqa
        api_post = {"is_state_changed": False}

        return api_post

    def assign_variable_action(self) -> dict:
        """Assigns a value to a given variable"""

        parameter_value = self.record.parameter1
        variable_id = self.record.variable1_id
        self.variables.set_variable_value_by_variable_id(
            variable_id, parameter_value, self.profile_id, self.profile_curr_state_id)
        assinged_variable_action = {"is_state_changed": self.got_response}
        return assinged_variable_action

    def increment_variable_action(self) -> dict:
        """Increments a value to a given variable by the amount of the given paramter1"""

        number_to_add = int(self.record.parameter1)
        variable_id = self.record.variable1_id
        current_variable_value = self.variables.get_variable_value_by_variable_id(
            variable_id, self.lang_code, self.profile_id)
        self.variables.set_variable_value_by_variable_id(variable_id, str(
            int(current_variable_value) + number_to_add), self.profile_id, self.profile_curr_state_id)
        incremented_variable_action = {"is_state_changed": self.got_response}
        return incremented_variable_action

    def decrement_variable_action(self) -> dict:
        """Increments a value to a given variable by the amount of the given paramter1"""

        if isinstance(self.record.parameter1, str) and not self.record.parameter1.isdigit():
            raise ValueError(f"Parameter1 must be a number (got {self.record.parameter1})")
        number_to_add = int(self.record.parameter1)
        variable_id = self.record.variable1_id
        current_variable_value = self.variables.get_variable_value_by_variable_id(
            variable_id, self.lang_code, self.profile_id)
        self.variables.set_variable_value_by_variable_id(variable_id, str(int(current_variable_value) - number_to_add),
                                                         self.profile_id, get_curr_state_id(self.profile_id))
        decremented_variable_action = {"is_state_changed": False}
        return decremented_variable_action

    # I have put this in remark right now because this action need to work with multiple profiles,
    # But right now this change makes it difficult to do that. Will work on it later.

    # def condition_action(self):
    #     cursor.execute("""SELECT * FROM dialog_workflow_state  WHERE parent_state_id = %s""", [record.curr_state_id])
    #     child_nodes = cursor.fetchall()
    #     """I am assuming that in these child records the varaible id must be the same id of the parent variable id,
    #     and the parameter1 value is the value of a profile_id from which I shall get the age"""
    #     for child in child_nodes:
    #         profile_id = child["parameter1"]
    #         child_age = (profiles_dict_class.get(profile_id)).get_variable_value_by_id(record.variable1_id)
    #         if child_age < record.result_figure_max and child_age > record.result_figure_min:
    #             profile_context.curr_state_id = child["next_state_id"]
    #             return True, None
    #     return False, None

    def menu_action(self) -> dict:
        """This action show a menu of options to the user for which he should choose one from it.
            the options we show are the records such that their parent id is the id of the current record.
            First part of the action is sending the user the options.
            Second part of the action is getting the chosen option (i.e. incoming_message) and dealing with it."""

        fields_to_select = ["parameter1", "next_state_id"]
        table_name = "dialog_workflow_state_view"
        values_from_where_to_select = (self.record.curr_state_id,)
        variables_from_where_to_select = ["parent_state_id"]
        child_nodes = get_child_nodes_of_current_state(
            fields_to_select, table_name, values_from_where_to_select, variables_from_where_to_select)
        self.record.parameter1 = self.record.parameter1 or ""
        # Adds the question and instructions to the accumulated_message to be sent to user.
        if not self.got_response:
            self.accumulated_message = self.accumulated_message + self.record.parameter1 + \
                                       f"{seperator}Please choose EXACTLY ONE option between 1-{len(child_nodes)}:{seperator}"
        is_state_changed, next_state_id, outgoing_message = self.generic_user_choice_action(
            record=self.record, accumulated_message=self.accumulated_message, child_nodes=child_nodes,
            choosed_exactly_one_option=True, got_response=self.got_response, chosen_numbers=self.incoming_message,
            profile_context=self.profile_context)
        if outgoing_message is not None:  # returns the outgoing message to send to the user.
            menu_action_selected = {"is_state_changed": False, "outgoing_message": outgoing_message}
            return menu_action_selected
        self.profile_curr_state_id = next_state_id
        self.accumulated_message = ""
        menu_action_selected = {"is_state_changed": is_state_changed}

        return menu_action_selected

    @staticmethod
    def age_detection() -> dict:
        # TODO: implament this function

        #     """Action that recieves a path to a picture (for now the picture has to be stored in the folder)
        #         and returns the approximate age of the person in the picture.
        #         Stores the picture in database storage."""
        #     if not self.got_response:
        #         self.accumulated_message += "Please insert a path to the picture" + seperator
        #         outgoing_message = process_message(communication_type= COMMUNICATION_TYPE, action_type= action_enum.AGE_DETECTION, message= self.accumulated_message) # noqa
        #         if COMMUNICATION_TYPE == communication_type_enum.WEBSOCKET:
        #             return False, outgoing_message
        #         else:
        #             self.accumulated_message = ""
        #             self.incoming_message = input()
        #     age_range = DetectAge.detect(self.incoming_message)
        #     self.accumulated_message += f'The approximate age of the picture you have sent is: {age_range}{seperator}'
        #     insert_profile_variable_value(self.profile_id, self.record.variable1_id, age_range, self.profile_curr_state_id)
        #     store_age_detection_picture(age_range, self.profile_curr_state_id)
        age_detected = {"is_state_changed": False}

        return age_detected

    def multi_choice_poll(self) -> dict:
        """ Similar to Menu Action. If the user chose a single option we jump to next_state_id of the chosen option.
            Otherwise, we save the answers and jump to the next_state_id of the parent."""

        fields_to_select = ["parameter1", "next_state_id"]
        table_name = "dialog_workflow_state_view"
        values_from_where_to_select = (self.record.curr_state_id,)
        variables_from_where_to_select = ["parent_state_id"]
        child_nodes = get_child_nodes_of_current_state(
            fields_to_select, table_name, values_from_where_to_select, variables_from_where_to_select)
        self.record.parameter1 = self.record.parameter1 or ""
        # Adds the question and instructions to the accumulated_message to be sent to user.
        if not self.got_response:
            self.accumulated_message = self.accumulated_message + self.record.parameter1 + seperator + \
                                       f"Please select your desired choices, You may select any of the numbers between 1-{len(child_nodes)} with a comma seperator between each choice:{seperator}" # noqa
        is_state_changed, next_state_id, outgoing_message = self.generic_user_choice_action(
            record=self.record, accumulated_message=self.accumulated_message, child_nodes=child_nodes,
            choosed_exactly_one_option=False, got_response=self.got_response, chosen_numbers=self.incoming_message,
            profile_context=self.profile_context)
        if outgoing_message is not None:  # returns the outgoing message to send to the user.
            choises = {"is_state_changed": False, "outgoing_message": outgoing_message}
        else:
            self.profile_curr_state_id = next_state_id
            choises = {"is_state_changed": is_state_changed}

        return choises

    def present_child_groups_names_by_id(self) -> dict:
        """Presents all the groups that their parent id is the given one. Does so recursively"""

        child_groups = Group(int(self.record.parameter1))
        self.accumulated_message = "Here are the interests:" + seperator
        groups = child_groups.get_child_group_names()
        for i in range(len(groups)):
            self.accumulated_message += groups[i] + seperator + str(i) + "\n"
        child_groups_names_by_id = {"is_state_changed": True, "outgoing_message": self.accumulated_message}
        return child_groups_names_by_id

    def present_groups_with_certain_text(self) -> dict:
        """Present all groups that their text contains the given text. (e.g: given text: 'sport' -> 'sports', 'walking sport'...).
            Saves the chosen options in profile context."""

        groups = self.get_groups_with_text(self.record.parameter1)
        groups_with_certain_text = {"is_state_changed": False}
        if not self.got_response:
            self.accumulated_message += "Please choose your desired interests. You may select more than one choice with a comma seperator." + seperator # noqa
            for i, child in enumerate(groups):
                self.accumulated_message = self.accumulated_message + f'{i + 1}) {child["title"]}{seperator}'

            outgoing_message = process_message(
                communication_type=COMMUNICATION_TYPE,
                action_type=WorkflowActionEnum.PRESENT_GROUPS_WITH_CERTAIN_TEXT,
                message=self.accumulated_message)
            if COMMUNICATION_TYPE == CommunicationTypeEnum.WEBSOCKET:
                groups_with_certain_text = {"is_state_changed": False, "outgoing_message": outgoing_message}
            else:
                self.accumulated_message = ""
                # TODO: chosen_numbers = input()
        else:
            chosen_numbers = self.incoming_message.split(',')
            chosen_numbers_list = [int(x) for x in chosen_numbers]
            self.profile_context.groups.extend(
                [groups[chosen_number] for chosen_number in chosen_numbers_list])

        return groups_with_certain_text

    def insert_missing_data(self) -> dict:
        """Asks the user for missing data (e.g. please insert your first name), and after getting a response,
            inserts the given value to the relevant table to fill the missing data.
            The record, field name, table and scehma in which the data should be inserted into are given in parameter1 as:
            <schema>,<table>,<field name>,<record id> (e.g. user,user_table,first_name,1)"""

        parameter1_list = self.record.parameter1.split(",")
        schema = parameter1_list[0]
        table = parameter1_list[1]
        field_name = parameter1_list[2]
        record_id = parameter1_list[3]
        if not self.got_response:
            self.accumulated_message += f"Please insert your {field_name}{seperator}"
            outgoing_message = process_message(
                communication_type=COMMUNICATION_TYPE, action_type=WorkflowActionEnum.TEXT_MESSAGE_ACTION,
                message=self.accumulated_message)
            if COMMUNICATION_TYPE == CommunicationTypeEnum.WEBSOCKET:
                missing_data = {"is_state_changed": False, "outgoing_message": outgoing_message}

                return missing_data
            else:
                self.accumulated_message = ""
                self.incoming_message = input(self.accumulated_message + " ")
        else:
            try:
                connection = Connector.connect(schema)
                cursor = connection.cursor(dictionary=True, buffered=True)
                # cursor.execute(f"""USE {schema}""")
                cursor.execute(
                    f"""UPDATE {table} SET {field_name} = '{self.incoming_message}' WHERE (id= {record_id})""")
                # cursor.execute("""USE dialog_workflow""")
                connection.commit()
            except Exception as exception:  # If one of the arguments isn't valid
                self.logger.error("Invalid parameter1", object=exception)
        missing_data = {"is_state_changed": False}

        return missing_data

    def present_and_choose_script(self) -> dict:
        """Action for asking the user which workflo script he would like to run next and change the next state id according to his choice."""

        connection = Connector.connect('dialog_workflow')
        cursor = connection.cursor(dictionary=True, buffered=True)
        cursor.execute(
            """SELECT d.start_state_id, dml.title FROM dialog_workflow_script_view AS d JOIN dialog_workflow_script_ml_view AS dml on dml.dialog_workflow_script_id=d.dialog_workflow_script_id WHERE dml.lang_code = %s OR dml.lang_code IS NULL""", # noqa
            (self.lang_code.value,))
        available_scripts_dict = cursor.fetchall()
        # TODO: prefer not null lang_code if exist
        available_scripts = [script["title"] for script in available_scripts_dict]
        outgoing_message = "Please choose your desired script out of the following:" + seperator
        menu = generic_menu(options=available_scripts, got_response=self.got_response,
                            chosen_numbers=self.incoming_message,
                            choose_one_option=True, outgoing_message=outgoing_message)
        if COMMUNICATION_TYPE == CommunicationTypeEnum.WEBSOCKET and not self.got_response:
            present_and_choosed_script = {"is_state_changed": False, "outgoing_message": menu}
        else:
            self.profile_curr_state_id = available_scripts_dict[menu[0] - 1]["start_state_id"]
            present_and_choosed_script = {"is_state_changed": True}

        return present_and_choosed_script

    def present_form_action(self) -> dict:
        form_id = self.record.parameter1
        if not isinstance(form_id, int):
            if isinstance(form_id, str) and form_id.isdigit():
                form_id = int(form_id)
            else:
                raise ValueError(f"parameter1 must be an integer (got {form_id})")
        save_additional_parameters = {"from_subsystem_id": 2,  # TODO do not use hard coded
                                      "channel_id": 14,
                                      "last_dialog_workflow_state_id": self.record.curr_state_id, }
        outgoing_messae_dict = CompoundMessage(
            form_id=form_id, save_additional_parameters=save_additional_parameters).get_compound_message_dict()
        form_actions = {"is_state_changed": False, "outgoing_message": outgoing_messae_dict}
        return form_actions

    def generic_user_choice_action(self, *, record: DialogWorkflowRecord, accumulated_message: str, child_nodes: list,
                                   choosed_exactly_one_option: bool, got_response: bool, chosen_numbers: str,
                                   profile_context: ProfileContext) -> (bool, int, str):  # type: ignore # TODO: return dict
        """Sends the user a question with a couple of answers. This function is generic and can let the user choose either
            exactly one option, or more than one. Each case is handeled differently.
            Returns:
                1. True if the users' next state should be changed to a child next state, False if there's no need to change it's state
                2. The next state id that the profile_context should be in, or None if the action didn't result in a change of state.
                3. The outging message to be sent to the user, or None if the message had already been sent."""
        # This is the first part of the action: sending the request to the user and waiting for an answer.
        is_state_changed = False
        if not got_response:
            for i, child in enumerate(child_nodes):
                accumulated_message = accumulated_message + \
                                      f'{i + 1}) {child["parameter1"]}{seperator}'
            outgoing_message = process_message(communication_type=COMMUNICATION_TYPE,
                                               action_type=WorkflowActionEnum.TEXT_MESSAGE_ACTION,
                                               message=accumulated_message)
            if COMMUNICATION_TYPE == CommunicationTypeEnum.WEBSOCKET:
                next_state_id = record.next_state_id
                return is_state_changed, next_state_id, outgoing_message
            elif not chosen_numbers:
                chosen_numbers = input("Please choose your desired option:")

        # The user has to pick exactly one option
        if choosed_exactly_one_option and chosen_numbers.isdigit():
            if int(chosen_numbers) > len(child_nodes) or int(chosen_numbers) < 1:
                raise ValueError(
                    f"Invalid input: {chosen_numbers}, (expected an integer between 1 and {len(child_nodes)})")
            profile_next_state = (
                child_nodes[int(chosen_numbers) - 1])["next_state_id"]
            is_state_changed = True
            next_state_id = profile_next_state
        # In this case the user can choose more than one option:
        else:
            chosen_numbers = chosen_numbers.split(',')
            try:
                chosen_numbers_list = [int(x) for x in chosen_numbers if x.isdigit()]
                if not all(x.isdigit() for x in chosen_numbers):
                    self.logger.warning(
                        f"Ignored non-integers from: `{chosen_numbers}`, (expected a comma separated list of integers)")
            except ValueError:
                raise ValueError(
                    f"Invalid input: {chosen_numbers}, (expected an integer or a comma separated list of integers)")

            # If he chooses more than one, we store the chosen options and jump to the next_state_id of the parent.
            list_of_options = [option["parameter1"] for option in child_nodes]
            profile_context.save_chosen_options(
                question_asked=record.parameter1, variable_id=record.variable1_id,
                chosen_numbers_list=chosen_numbers_list, list_of_options=list_of_options)
            next_state_id = record.next_state_id

        outgoing_message = None
        return is_state_changed, next_state_id, outgoing_message

    def display_message_template_action(self) -> dict:
        # call compound message
        if not self.record.message_template_id:
            raise ValueError(f"message_template_id is empty for record {self.record}")
        recipient = Recipient(profile_id=self.profile_id, preferred_lang_code_str=self.lang_code.value,
                              user_id=self.user_context.get_effective_user_id(),
                              first_name=self.user_context.get_real_name())

        outgoing_message = CompoundMessage(message_template_id=self.record.message_template_id,
                                           recipients=[recipient]).get_compound_message_dict()
        display_message_template = {"is_state_changed": False, "outgoing_message": outgoing_message}
        return display_message_template

    @staticmethod  # TODO move to another repository
    def get_groups_with_text(text: str) -> list:
        connection = Connector.connect('group')
        cursor = connection.cursor(dictionary=True, buffered=True)
        # cursor.execute("""USE `group`""")
        cursor.execute(
            f"""SELECT title, group_id FROM group_ml_table WHERE title LIKE '%{text}%'""")
        groups_with_text = cursor.fetchall()

        return groups_with_text
